import os
import jwt 

from flask import Flask, request, session
from flask_session import Session
from FlaskOIDC import FlaskOIDC

from config import oidc_config, session_config

DEBUG = os.environ.get('DEBUG', False)

app = Flask(__name__)
app.config.from_mapping(session_config)

Session(app)

auth = FlaskOIDC(config=oidc_config, app=app)

#
# Authentication: require_login decorator
#
@app.route('/login')
@app.route('/logon')
@auth.require_login
def login():
    """ Login """
    
    return {'username': session['username']}

#
# logout 
#
@app.route('/logout')
@app.route('/logoff')
def bye():
    """ Logout """

    if auth.is_authenticated:
        return auth.initiate_logout(next=request.url)
    else:
        return 'ok'

#
# Authorization: require_user decorator
#
@app.route('/bob')
@auth.require_user(['bob','rkras'])
def bob():
    """ Only bob """
    return 'hi bob'

#
# Authorization: require_attribute decorator
#
@app.route('/sa')
@auth.require_attribute('groups', ['netadmin', 'sysadmin'])
def sa():
    return 'hi sa'


if DEBUG:
    # DEBUG helpers

    @app.route('/mysession')
    def my_session_status():
        """ Dump username/attributes from OIDC id token state """
        
        return {
            'username': auth.my_username,
            'attributes': auth.my_attrs
        }


    @app.route('/<scpname>/expire')
    @app.route('/expire')
    @auth.assert_login
    def expire(scpname=auth.token_name):
        """ Hasten token expiration by 1 hr. - for testing auto token refresh """

        session[scpname]['exp'] -= 3600
        session.modified = True
        return 'ok'


    def decode_jwt(token):
        """ decode and return token response. """

        payload = jwt.decode(token, options={'verify_signature': False})
        return payload


    @app.route('/<scpname>/new')   
    @auth.assert_login
    def new_scope_token(scpname='meh'):
        """ 
        Request new tokens with new scope.
        
        New requested scope is provided a query string "scp="
        
            GET /dbaccess/new?scp=[db.dbread]
        
        'dbaccess' is the name then used to access the tokens.
        """
        newscp = request.args.get('scp')
        tokens = auth.get_access_token(scpname, scope=newscp)
        return {'tokens': tokens}


    @app.route('/id')          # decode id token for login scope
    @app.route('/<scpname>/id')    # decode id token for named scope
    @auth.assert_login
    def decode_id_tok(scpname=auth.token_name):
        """ 
        decode id token 
        
            GET /id
        
        returns id token for login (default) scope

            GET /dbaccess/id
        
        returns id token for `dbaccess` named tokens
        """

        return decode_jwt(session[scpname]['id_token'])


    @app.route('/id/raw')
    @app.route('/<scpname>/id/raw')
    @auth.assert_login
    def raw_id_tok(scpname=auth.token_name):
        """ Return the id token raw """

        return session[scpname]['id_token']


    @app.route('/access')          # decode access token from login
    @app.route('/<scpname>/access')    # decode access token for named scope
    def decode_access_tok(scpname=auth.token_name):
        """ 
        decode access token 
        
            GET /access
        
        return access token for login (default) scope

            GET /dbaccess/access
        
        return access token for 'dbaccess' named token
        """

        return decode_jwt(session[scpname]['access_token'])


    @app.route('/access/raw')
    @app.route('/<scpname>/access/raw')
    @auth.assert_login
    def raw_access_tok(scpname=auth.token_name):
        """ Return the access token raw """

        return session[scpname]['access_token']




# Example configuration template
oidc_config = {
    "client_id": "IDP Provided Client ID",
    "client_secret": "IDP Provided Secret",
    "discovery_url": "https://login.microsoftonline.com/<tenentid>/V2.0/.well-known/openid-configuration",
    "client_scope": ["openid", "email", "profile", ],
    "user_attr": 'email'
}

session_config = {
    'secret_key': b'now is the time to test this',
    'SESSION_TYPE': 'filesystem', # You can use redis/memcache/etc - see Flask-Session docs
    'SESSION_FILE_DIR': './.cache', # for filesystem cache
    'PERMANENT_SESSION_LIFETIME' : 300,
    'SESSION_COOKIE_NAME' :'demo', # name of the cookie
    'SESSION_COOKIE_HTTPONLY': True,
    'SESSION_COOKIE_SECURE': True,
}


from flask import Response, abort

from time import time

now = lambda: int(time())

def response_nocache(status=200, body=''):
    
    response = Response(status=status, response=body)
    
    # MDN recommended for various browsers
    response.headers.add_header('Cache-Control', 'no-cache')
    response.headers.add_header('Cache-Control', 'must-revalidate')
    response.headers.add_header('Pragma', 'no-cache')
    response.headers.add_header('Expires', 'Sun, 25 Jul 2021 15:42:14 GMT')
    return response


UnauthorizedError = lambda body=None: abort(status=401)

BadRequestError = lambda body=None: abort(status=400)

AppError = lambda body=None: abort(status=400)

ConflictError = lambda body=None: abort(status=400)

ForbiddenError = lambda body=None: abort(status=403)




import requests as req
import sys

from .JWKS import Jwks


class   OidcDiscover:
    """
    OidcDiscover - discover OIDC configuration and JWK signing keys

    discov = OIDC_discover(url, timeout)
    
        Autodiscover OIDC endpoint from discovery url
        url - discovery url
        timeout - timeout for config retrival (def: 4 Seconds)
    """

    def __init__(self, url, timeout = 4):

        self.discovery_url = url
        self.add_offline_scp = False
        self.discovery_complete = False
        self.timeout = timeout

        self.discover()
        

    def discover(self, url=None):
        """Retrieve OIDC autodiscovery to configure authorizer."""
        
        if url is None:
            url = self.discovery_url

        print(f'** Initating OIDC autodiscovery on "{url}" **', file=sys.stderr)

        try:
            discovery_config = req.get(url, timeout=self.timeout).json()
            if 'error' in discovery_config:
                print(discovery_config['error_description'], file=sys.stderr)
                raise Exception(discovery_config['error_description'])

            self.issuer = discovery_config['issuer']
            self.auth_url = discovery_config['authorization_endpoint']
            self.token_url = discovery_config['token_endpoint']

            self.jwks_uri = discovery_config.get('jwks_uri')
            
            self.logout_url = discovery_config.get('end_session_endpoint')
            self.scopes_supported = discovery_config.get('scopes_supported',[])
            
        except Exception as e:

            print(f'** OIDC Autodiscovery FAILED using "{url}" **',file=sys.stderr)
            raise e

        self.jwks = Jwks(self.jwks_uri)

        print(f'** OIDC autodiscovery completed.**', file=sys.stderr)




"""
Jwks - retrieve public keys from url and decode/verify tokens

"""
import base64
import json
import jwt
import requests as req
import sys

from cryptography.hazmat.primitives import hashes
#from cryptography.hazmat.primitives.asymmetric import padding


def __base64url_decode(input):
    """pad out input length to mod 4 and decode."""

    if isinstance(input, str):
        input = input.encode("ascii")

    rem = len(input) % 4

    if rem > 0:
        input += b"=" * (4 - rem)

    return base64.urlsafe_b64decode(input)


class   Jwks:

    def __init__(self, url=None, timeout=4):
        """
        jwks = Jwks(url, timeout)

        Keys are obtained from url on object instantiation
            url - points to jwks json object
            timeout - for retrieving url (def 4 sec)

        """

        self.url = url
        self.timeout = timeout
        self.pub_keys = {}

        if url:
            self.pub_keys = self._load_jwks(url)
        else:
            self.pub_keys = {}

        self.rsa = jwt.algorithms.RSAAlgorithm(hashes.SHA256)


    def load_jwks(self, url):

        self.pub_keys = self._load_jwks(url)

        
    def _load_jwks(self, url):
        """ Load keys from public endpoint. """

        print(f'*** Loading JWKS from {url}', file=sys.stderr)
        pub_keys = {}
        try:
            jwks_config = req.get(url, timeout=self.timeout).json()

            if 'keys' not in jwks_config:
                raise Exception('no keys in jwks')

            for jwk in jwks_config['keys']:
                # Load only RSA signature keys
                if jwk['kty'] == 'RSA' and jwk['use'] == 'sig':
                    pub_keys[jwk['kid']] = jwt.algorithms.RSAAlgorithm.from_jwk(json.dumps(jwk))

        except Exception as e:
            print(f'Failure loading jwks {str(e)}', file=sys.stderr)
            print(f'*** Loading JWKS Falied. ***', file=sys.stderr)
            raise e
        
        print(f'*** JWKS loading completed: {len(pub_keys)} keys loaded. ***', file=sys.stderr)
        return pub_keys


    def decode(self, token, options={}, **kwargs):
        """
        jwks.decode(token, **kwargs)

            Decodes token with PyJWT perameters/options.
            Returns payload on success, None on failure
        """

        try:
            kid = jwt.get_unverified_header(token)['kid']
            
            if self.pub_keys and kid in self.pub_keys:
                pub_key = self.pub_keys[kid]
                try:
                    
                    return jwt.decode(token, key=pub_key, algorithms=['RS256'], options=options, **kwargs)
                    
                except Exception as e:
                    print(f'Error: verify/decode Failed {str(e)}', file=sys.stderr)
                    return None
            else:
                print(f'Error: can not verify: Public Key with KID {kid} has not been loaded', file=sys.stderr)
        
        except Exception as e:
            print(f'Error: decode Bad Token: {str(e)}', file=sys.stderr)

        return None
    

    def signature_verify(self, token):
        """
        jwks.signature_verify(token)

            Verify signature only (not iat, issuer, etc.)
            Returns payload on success, None on failure
        """

        try:
            kid = jwt.get_unverified_header(token)['kid']
            
            if self.pub_keys and kid in self.pub_keys:
                pub_key = self.pub_keys[kid]
                try:
                    return self._jwt_verify_signature(pub_key, token)

                except Exception as e:
                    print(f'Error: signature_verify Failed: {str(e)}', file=sys.stderr)
            else:
                print(f'Error: signature_verify: No Public Key with KID: {kid}', file=sys.stderr)
        
        except Exception as e:
            print(f'Error: signature_verify Bad Token: {str(e)}', file=sys.stderr)
    
        return None


    def decode_noverify(self, token, **kwargs):
        """    
        jwks.decode_noverify(token)

            Decodes token, but does no verification. 
            Returns payload on success, None on failure
        """

        try:
            return jwt.decode(token, options={'verify_signature': False, **kwargs})

        except Exception as e:
            print(f'Error: decode_noverify: decode failed {str(e)}', file=sys.stderr)
            return None


    def _jwt_verify_signature(self, pub_key, token):
        """ Verify signing on token with public key. """

        # split token
        (message, signature) = token.rsplit('.', 1)
        (header, payload) = message.split('.', 1)
        
        # verify
        self.rsa.verify(message.encode('utf-8'), pub_key, __base64url_decode(signature))
        
        payload_data = json.loads(__base64url_decode(payload))

        return payload_data




import json

from cryptography.fernet import Fernet

class OIDCstate:
    """
    Serializes (json), Fernet encrypts, decrypts, and deserializes a dict
    """

    def __init__(self, key=None, ttl=60):

        self.key = key if key else Fernet.generate_key()
        self.ttl = ttl
        self.f = Fernet(self.key)

    
    def serial(self,state):
        """
        Serialize
            - json serialize state
            - Fernet encrypt and return as string
        """

        bdat = json.dumps(state).encode('utf-8')
        return self.f.encrypt(bdat).decode()


    def deserial(self, bdat):
        """ 
        Deserialize 
            - Fernet decrypt and validate TTL against replay
            - deserialize json
        """
        if type(bdat) is str:
            bdat = bdat.encode('utf-8')

        jdat = self.f.decrypt(bdat, ttl=self.ttl)
        dat = json.loads(jdat)

        return dat



import time
from urllib.parse import urlencode

import requests as req
from flask import (
        Blueprint, 
        request, 
        session, 
        current_app, 
        redirect, 
        url_for
    )

from FlaskOIDC.oidc_discover import OidcDiscover
from FlaskOIDC.oidc_state import OIDCstate
from FlaskOIDC.flask_utils import (
        UnauthorizedError, 
        BadRequestError, 
        ConflictError
    )

default_oidc_scope = ['openid', 'email', 'profile']
default_user_attr = 'email'

now = lambda : int(time.time())


class FlaskOIDC(OidcDiscover,Blueprint):
    """
    auth = BottleOIDC(app, config, ...)

    OIDC Service Provider for Bottle
    Uses Authorization Code Grant flow

    """

    def __init__( self,
            config,
            sess_username = 'username',
            sess_attr = 'oidc_attr',
            app=None,
        ):

        self.client_id = config['client_id']
        self.client_secret = config['client_secret']
        self.scopes = config.get('client_scope', default_oidc_scope)
        self.username_id = config.get('user_attr', default_user_attr)

        self.token_name = 'oidc_tokens'
        self.sess_username = sess_username
        self.sess_attr = sess_attr

        # autodiscovery of oidc config in base class
        discovery_url = config['discovery_url']
        timeout = config.get('timeout', 4)      # undocumented
        super().__init__(discovery_url, timeout=timeout)

        # msft special - 'offline_access' provides refresh tokens
        if 'offline_access' in self.scopes_supported:
            self.scopes.append('offline_access')
        
        # initialize state creator # state_key and state_ttl undocumented
        self.state = OIDCstate(key=config.get('state_key'), ttl=config.get('state_ttl',60))
        
        # make this a blueprint
        Blueprint.__init__(self, name='oidcsp', import_name=__name__)

        # OIDC authorized - receives code grant redirect form IdP via client
        self.add_url_rule(
            '/oidc/authorized',
            endpoint='authorized',
            view_func=self._finish_oauth_login
        )
        
        self.login_hooks = [self._id_token_hook]

        if not config.get('logout_idp',False):
            # Local logout only (i.e. don't notify IdP)
            self.logout_url = None
        
        if app:
            # app was specified - install ourself as a blueprint
            app.register_blueprint(self,)


    @property
    def is_authenticated(self):
        """ True if user has authenticated. """

        return self.sess_username in session and session[self.sess_username]


    @property
    def my_username(self):
        """ Return username for the current session. """

        return session[self.sess_username] if self.is_authenticated else None
    

    @property
    def my_attrs(self):
        """ Return collected assertions for the current session. """

        return session[self.sess_attr] if self.is_authenticated else {}
    

    def initiate_login(self, next=None, scopes=None, **kwargs):
        """ Initiate an OIDC/Oauth2 login. (return a redirect.) """

        # 'next' url - return to this after tokens acquired.
        state = {
            'next': next if next else request.params.get('next','/')
        }
        
        params = {
            'client_id' : self.client_id,
            'response_type' : 'code',
            'redirect_uri' : url_for('oidcsp.authorized', _external=True),
            'response_mode': 'query',
            'scope' : ' '.join(scopes if scopes else self.scopes),
            'state' : self.state.serial(state),
        }
        
        # These are microsoft Azure AD login extentensions
        if request.args.get('login_hint'):
            params.update({'login_hint': request.args.get('login_hint')})
        
        if kwargs.get('userhint'):
            # priority over any in request query string
            params.update({'login_hint': kwargs.get('userhint')})
        
        if request.args.get('domain_hint'):
            params.update({'domain_hint': request.args.get('domain_hint')})

        if request.args.get('prompt'):
            params.update({'prompt': request.args.get('prompt')})
        
        if kwargs.get('force_reauth'):
            params.update({'prompt':'login'})

        return redirect(self.auth_url + '?' + urlencode(params))


    # route: /authorized
    def _finish_oauth_login(self):
        """ Callback Route: Complete login by obtaining id and access tokens. """
        
        if 'error' in request.args:
            msg = f'OIDC: AuthNZ error: {request.args.get("error_description")}'
            current_app.logger.info(msg)
            return BadRequestError(msg)

        try:
            # Validate and deserialize state
            state = self.state.deserial(request.args.get('state'))

        except Exception as e:
            msg = 'OIDC: Authentication request was not outstanding'
            current_app.logger.info(msg, str(e))
            return BadRequestError(msg)

        code = request.args.get('code')

        # Prepare to exchange code for tokens
        params = {
            'client_id' : self.client_id,
            'client_secret' : self.client_secret,
            'grant_type' : 'authorization_code',
            'code': code,
            'redirect_uri' : url_for('oidcsp.authorized', _external=True),
        }
        
        try:
            current_app.logger.debug(f'OIDC: exchanging code {code[:10]}...{code[-10:]} for tokens')
            resp = req.post(self.token_url, data=params, timeout=self.timeout)
            tokens = resp.json()
            
            if 'error' in tokens:
                msg = f'OIDC: error exchanging code for tokens: {tokens["error_description"]}'
                current_app.logger.info(msg)
                return ConflictError(msg)
            
            try:
                # authenticate and decode the id token
                idtok = self.jwks.decode(tokens['id_token'], audience=self.client_id)
                
                tokens['exp'] = idtok['exp']
                session[self.token_name] = tokens

                username = idtok.get(self.username_id, 'Authenticated User')       
                attrs = idtok

                # Run all login hooks
                for login_hook in self.login_hooks:
                    username, attrs = login_hook(username, attrs)

                attrs.update({
                    'authenticated' : now()
                })
                
                current_app.logger.info(f'OIDC: User "{username}" authenticated')
                session[self.sess_attr] = attrs
                session[self.sess_username] = username

            except Exception as e:
                current_app.logger.info(f'Error: OIDC: failed to verify token: {str(e)}')
                return UnauthorizedError('OIDC: failed to verify id token')

        except Exception as e:
            current_app.logger.info(f'Error: OIDC: token acquisition failed: {str(e)}')
            return UnauthorizedError('OIDC: Error acquiring id token')

        if 'next' in state:
            return redirect(state['next'])
        else:
            return f'OIDC: authenticated "{username}"'


    def initiate_logout(self, next=None):
        """ Clear session and redirect to provider logout. """

        if next is None:
            next = request.args.get('next') 

        if self.is_authenticated:
            user = self.my_username
        else:
            user = 'Anonymous'
        
        current_app.logger.info(f'OIDC: user "{user}" logged out')

        # since we did the authentication, we should do this:
        session.clear()

        if self.logout_url and next:
            return redirect(self.logout_url +'?' + urlencode({'post_logout_redirect_uri': next}))

        elif self.logout_url:
            return redirect(self.logout_url)

        elif next:
            return redirect(next)

        else:
            return 'Logout complete'


    def _token_expire_check(self, token_name=None):
        """ Refresh token if needed. """
        
        if not token_name:
            # default is the base authenticator tokens
            token_name = self.token_name
        
        if now() < session[token_name]['exp']:
            # The tokens are still valid
            return True

        current_app.logger.debug(f'OIDC: Auto-refreshing expired "{token_name}" token')

        tokens = self._get_token_with_refresh(token_name)

        if tokens:
            idtok = self.jwks.decode(tokens['id_token'], options={'verify_signature':False})

            tokens['exp'] = idtok['exp']
            session[token_name] = tokens

            current_app.logger.debug(f'OIDC: Token refreshed')
            return True
        
        else:
            current_app.logger.info(f'OIDC: session token refresh for "{token_name}" failed.')
            return False


    def _get_token_with_refresh(self, token_name=None, scope=None):
        """ Get a new tokens using the refresh token. """
        
        if not token_name:
            # default is the base authenticator tokens
            token_name = self.token_name
        
        if token_name in session:
            current_tokens = session[token_name]
        else:
            # this is a new token_name, use the oidc tokens for refresh
            current_tokens = session[self.token_name]
        
        if 'refresh_token' not in current_tokens:
            # we don't have a refresh token to use
            return None

        params = {
            'client_id' : self.client_id,
            'client_secret' : self.client_secret,
            'grant_type' : 'refresh_token',
            'refresh_token' : current_tokens['refresh_token'],
        }

        if scope:
            # specific scope is requested
            params.update({'scope': scope})
        
        resp = req.post(self.token_url, data=params)
        new_tokens = resp.json()
        
        if 'error' in new_tokens:
            # There was a failure
            current_app.logger.debug(f'OIDC: Error: refreshing tokens: {new_tokens["error_description"]}')
            return None
        
        idtok = self.jwks.decode(new_tokens['id_token'], options={'verify_signature' :False})
        
        new_tokens['exp'] = idtok['exp']

        return new_tokens


    def _id_token_hook(self, user, attr):
        """ Remove unneeded id_token data from session attributes """

        for key in ['aud', 'iss', 'iat', 'nbf', 'exp', 'aio', 'tid','uti', 'ver', 'wids']:
            if key in attr: del attr[key]

        # username part of email:
        user = user.split('@')[0]

        # Add username as an attribute as well   
        attr['username'] = user

        return user, attr


    def get_access_token(self, token_name=None, scope=None):
        """ Get and cache an access_token for given scopes. """
        
        if not token_name:
            # default is the base authenticator tokens
            token_name = self.token_name

        if token_name in session and session[token_name]['exp'] < now():
            # this token is expired - remove it
            del session[token_name]

        if token_name in session:
            # return the current cached token
            return session[token_name]

        else:
            # nothing cached, get a new token
            new_tokens = self._get_token_with_refresh(scope=scope)

            if new_tokens:
                # token is valid, so save it
                session[token_name] = new_tokens
                return new_tokens
            else:
                # no token provided - just to be explicit
                return None
    

    # api: @auth.assert_login decorator
    def assert_login(self, f):
        """ Return error on view if user is not authenticated """

        def _wrapper(*args, **kwargs):

            if self.is_authenticated and self.token_name in session:
                
                if self._token_expire_check(self.token_name):

                    return f(*args, **kwargs)

            # either no user in this session or a refresh failed - full login...
            return UnauthorizedError()

        _wrapper.__name__ = f.__name__
        return _wrapper       

    # api: @auth.require_login decorator.
    def require_login(self, f):
        """ Decorator for forcing authenticated. """

        def _wrapper(*args, **kwargs):

            if self.is_authenticated and self.token_name in session:
                
                if self._token_expire_check(self.token_name):

                    return f(*args, **kwargs)

            # either no user in this session or a refresh failed - full login...
            return self.initiate_login(next = request.url)

        _wrapper.__name__ = f.__name__
        return _wrapper
    

    def add_login_hook(self,f):
        """ Decorator for adding login hook. """

        self.login_hooks.append(f)           
        return f


    def require_user(self, user_list):
        """ Decorator passes on specific list of usernames. """

        def _outer_wrapper(f):

            def _wrapper(*args, **kwargs):
                if self.my_username in user_list:
                    return f(*args, **kwargs)
                
                return UnauthorizedError('Not Authorized')
            
            _wrapper.__name__ = f.__name__
            return _wrapper
        
        return _outer_wrapper
                

    def require_attribute(self, attr, value):
        """ Decorator requires specific attribute value. """

        def test_attrs(challenge, standard):
            """Compare list or val the standard."""

            stand_list = standard if type(standard) is list else [standard]
            chal_list = challenge if type(challenge) is list else [challenge]

            for chal in chal_list:
                if chal in stand_list:
                    return True
            return False
        
        def _outer_wrapper(f):

            def _wrapper(*args, **kwargs):

                if attr in self.my_attrs:
                    resource = session[self.sess_attr][attr]
                    
                    if test_attrs(resource, value):
                            return f(*args, **kwargs)

                return UnauthorizedError('Not Authorized')

            _wrapper.__name__ = f.__name__
            return _wrapper

        return _outer_wrapper
